/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%                                                                             %
%                  W   W  IIIII  DDDD    GGGG  EEEEE  TTTTT                   %
%                  W   W    I    D   D  G      E        T                     %
%                  W W W    I    D   D  G  GG  EEE      T                     %
%                  WW WW    I    D   D  G   G  E        T                     %
%                  W   W  IIIII  DDDD    GGGG  EEEEE    T                     %
%                                                                             %
%                X11 User Interface Routines for ImageMagick.                 %
%                                                                             %
%                                                                             %
%                              Software Design                                %
%                                John Cristy                                  %
%                              September 1993                                 %
%                                                                             %
%                                                                             %
%  Copyright 1994 E. I. du Pont de Nemours & Company                          %
%                                                                             %
%  Permission to use, copy, modify, distribute, and sell this software and    %
%  its documentation for any purpose is hereby granted without fee,           %
%  provided that the above Copyright notice appear in all copies and that     %
%  both that Copyright notice and this permission notice appear in            %
%  supporting documentation, and that the name of E. I. du Pont de Nemours    %
%  & Company not be used in advertising or publicity pertaining to            %
%  distribution of the software without specific, written prior               %
%  permission.  E. I. du Pont de Nemours & Company makes no representations   %
%  about the suitability of this software for any purpose.  It is provided    %
%  "as is" without express or implied warranty.                               %
%                                                                             %
%  E. I. du Pont de Nemours & Company disclaims all warranties with regard    %
%  to this software, including all implied warranties of merchantability      %
%  and fitness, in no event shall E. I. du Pont de Nemours & Company be       %
%  liable for any special, indirect or consequential damages or any           %
%  damages whatsoever resulting from loss of use, data or profits, whether    %
%  in an action of contract, negligence or other tortuous action, arising     %
%  out of or in connection with the use or performance of this software.      %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
*/

/*
  Include declarations.
*/
#include "magick.h"
#include "image.h"
#include "utility.h"
#include "X.h"
#include "widget.h"

/*
  Define declarations.
*/
#define MatteIsActive(matte_info,position)  \
  ((position.x >= (matte_info.x-matte_info.bevel_width)) && \
   (position.y >= (matte_info.y-matte_info.bevel_width)) &&  \
   (position.x < (matte_info.x+matte_info.width+matte_info.bevel_width)) &&  \
   (position.y < (matte_info.y+matte_info.height+matte_info.bevel_width)))
/*
  State declarations.
*/
#define ControlState  0x0001
#define DefaultState  0x0000
#define ExitState  0x0002
#define RedrawFileInfoState  0x0004
#define RedrawWidgetState  0x0008
#define UpdateFilelistState  0x0010

/*
  Forward declarations.
*/
static void
  XEditText _Declare((Display *,XWidgetInfo *,KeySym,char *,unsigned long)),
  XGetWidgetInfo _Declare((char *,XWidgetInfo *)),
  XHighlightWidget _Declare((Display *,XWindowInfo *,int,int)),
  XSetBevelColor _Declare((Display *,XWindowInfo *,unsigned int)),
  XSetMatteColor _Declare((Display *,XWindowInfo *,unsigned int)),
  XSetTextColor _Declare((Display *,XWindowInfo *,unsigned int));

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X D r a w B e v e l e d B u t t o n                                       %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XDrawBeveledButton draws a button with a highlighted upper and
%  left bevel and a shadowed lower and right bevel.  The highlighted and
%  shadowed bevels create a 3-D effect.
%
%  The format of the XDrawBeveledButton function is:
%
%      XDrawBeveledButton(display,window_info,button_info)
%
%  A description of each parameter follows:
%
%    o display: Specifies a pointer to the Display structure;  returned from
%      XOpenDisplay.
%
%    o window_info: Specifies a pointer to a X11 XWindowInfo structure.
%
%    o button_info: Specifies a pointer to a XWidgetInfo structure.  It
%      contains the extents of the button.
%
%
*/
static void XDrawBeveledButton(display,window_info,button_info)
Display
  *display;

XWindowInfo
  *window_info;

XWidgetInfo
  *button_info;
{
  int
    length,
    x,
    x1,
    x2,
    y,
    y1,
    y2;

  unsigned int
    bevel_width,
    width;

  XFontStruct
    *font_info;

  XPoint
    points[6];

  /*
    Draw upper and left beveled border.
  */
  x1=button_info->x;
  y1=button_info->y+button_info->height;
  x2=button_info->x+button_info->width;
  y2=button_info->y;
  bevel_width=button_info->bevel_width;
  points[0].x=x1;
  points[0].y=y1;
  points[1].x=x1;
  points[1].y=y2;
  points[2].x=x2;
  points[2].y=y2;
  points[3].x=x2+bevel_width;
  points[3].y=y2-bevel_width;
  points[4].x=x1-bevel_width;
  points[4].y=y2-bevel_width;
  points[5].x=x1-bevel_width;
  points[5].y=y1+bevel_width;
  XSetBevelColor(display,window_info,button_info->raised);
  XFillPolygon(display,window_info->id,window_info->widget_context,points,6,
    Complex,CoordModeOrigin);
  /*
    Draw lower and right beveled border.
  */
  points[0].x=x1;
  points[0].y=y1;
  points[1].x=x2;
  points[1].y=y1;
  points[2].x=x2;
  points[2].y=y2;
  points[3].x=x2+bevel_width;
  points[3].y=y2-bevel_width;
  points[4].x=x2+bevel_width;
  points[4].y=y1+bevel_width;
  points[5].x=x1-bevel_width;
  points[5].y=y1+bevel_width;
  XSetBevelColor(display,window_info,(unsigned int) !button_info->raised);
  XFillPolygon(display,window_info->id,window_info->widget_context,points,6,
    Complex,CoordModeOrigin);
  XSetFillStyle(display,window_info->widget_context,FillSolid);
  /*
    Draw matte.
  */
  XSetMatteColor(display,window_info,button_info->raised);
  XFillRectangle(display,window_info->id,window_info->widget_context,
    button_info->x,button_info->y,button_info->width,button_info->height);
  XSetTextColor(display,window_info,button_info->raised);
  x=button_info->x-bevel_width-1;
  y=button_info->y-bevel_width-1;
  if (button_info->raised || (window_info->depth == 1))
    XDrawRectangle(display,window_info->id,window_info->widget_context,x,y,
      button_info->width+(bevel_width << 1)+1,button_info->height+
      (bevel_width << 1)+1);
  if (button_info->text == (char *) NULL)
    return;
  /*
    Insure text will fit within the window.
  */
  font_info=window_info->font_info;
  for (length=strlen(button_info->text); length > 0; length--)
    if (XTextWidth(font_info,button_info->text,length) <= button_info->width)
      break;
  /*
    Draw text.
  */
  width=XTextWidth(font_info,button_info->text,length);
  x=button_info->x+(button_info->width >> 1)-(width >> 1);
  y=button_info->y+((button_info->height-
    (font_info->ascent+font_info->descent)) >> 1)+
    window_info->font_info->ascent;
  XDrawString(display,window_info->id,window_info->widget_context,x,y,
    button_info->text,length);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X D r a w B e v e l e d M a t t e                                         %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XDrawBeveledMatte draws a matte with a shadowed upper and
%  left bevel and a highlighted lower and right bevel.  The highlighted and
%  shadowed bevels create a 3-D effect.
%
%  The format of the XDrawBeveledMatte function is:
%
%      XDrawBeveledMatte(display,window_info,matte_info)
%
%  A description of each parameter follows:
%
%    o display: Specifies a pointer to the Display structure;  returned from
%      XOpenDisplay.
%
%    o window_info: Specifies a pointer to a X11 XWindowInfo structure.
%
%    o matte_info: Specifies a pointer to a XWidgetInfo structure.  It
%      contains the extents of the matte.
%
%
*/
static void XDrawBeveledMatte(display,window_info,matte_info)
Display
  *display;

XWindowInfo
  *window_info;

XWidgetInfo
  *matte_info;
{
  int
    x1,
    x2,
    y1,
    y2;

  unsigned int
    bevel_width;

  XPoint
    points[6];

  /*
    Draw upper and left beveled border.
  */
  x1=matte_info->x;
  y1=matte_info->y+matte_info->height;
  x2=matte_info->x+matte_info->width;
  y2=matte_info->y;
  bevel_width=matte_info->bevel_width;
  points[0].x=x1;
  points[0].y=y1;
  points[1].x=x1;
  points[1].y=y2;
  points[2].x=x2;
  points[2].y=y2;
  points[3].x=x2+bevel_width;
  points[3].y=y2-bevel_width;
  points[4].x=x1-bevel_width;
  points[4].y=y2-bevel_width;
  points[5].x=x1-bevel_width;
  points[5].y=y1+bevel_width;
  XSetBevelColor(display,window_info,False);
  XFillPolygon(display,window_info->id,window_info->widget_context,points,6,
    Complex,CoordModeOrigin);
  /*
    Draw lower and right beveled border.
  */
  points[0].x=x1;
  points[0].y=y1;
  points[1].x=x2;
  points[1].y=y1;
  points[2].x=x2;
  points[2].y=y2;
  points[3].x=x2+bevel_width;
  points[3].y=y2-bevel_width;
  points[4].x=x2+bevel_width;
  points[4].y=y1+bevel_width;
  points[5].x=x1-bevel_width;
  points[5].y=y1+bevel_width;
  XSetBevelColor(display,window_info,True);
  XFillPolygon(display,window_info->id,window_info->widget_context,points,6,
    Complex,CoordModeOrigin);
  XSetFillStyle(display,window_info->widget_context,FillSolid);
  /*
    Draw matte.
  */
  XFillRectangle(display,window_info->id,window_info->highlight_context,
    matte_info->x,matte_info->y,matte_info->width,matte_info->height);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X D r a w M a t t e T e x t                                               %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XDrawMatteText draws a matte with text.  If the text exceeds the
%  extents of the text, a portion of the text relative to the cursor is
%  displayed.
%
%  The format of the XDrawMatteText function is:
%
%      XDrawMatteText(display,window_info,text_info)
%
%  A description of each parameter follows:
%
%    o display: Specifies a pointer to the Display structure;  returned from
%      XOpenDisplay.
%
%    o window_info: Specifies a pointer to a X11 XWindowInfo structure.
%
%    o text_info: Specifies a pointer to a XWidgetInfo structure.  It
%      contains the extents of the text.
%
%
*/
static void XDrawMatteText(display,window_info,text_info)
Display
  *display;

XWindowInfo
  *window_info;

XWidgetInfo
  *text_info;
{
  char
    *text;

  int
    x,
    y;

  register int
    i;

  unsigned int
    width;

  XFontStruct
    *font_info;

  /*
    Determine beginning of the visible text.
  */
  font_info=window_info->font_info;
  width=text_info->width-font_info->max_bounds.width;
  if (text_info->cursor < text_info->marker)
    text_info->marker=text_info->cursor;
  else
    {
      text=text_info->marker;
      if (XTextWidth(font_info,text,text_info->cursor-text) > width)
        {
          text=text_info->text;
          for (i=0; i < (int) strlen(text); i++)
            if (XTextWidth(font_info,text+i,text_info->cursor-text-i) <= width)
              break;
          text_info->marker=text+i;
        }
    }
  /*
    Determine end of the visible text.
  */
  for (i=0; i < (int) strlen(text_info->marker); i++)
    if (XTextWidth(font_info,text_info->marker,i+1) > width)
      break;
  /*
    Draw text and text.
  */
  x=text_info->x+(font_info->max_bounds.width >> 1);
  y=text_info->y+font_info->ascent+(text_info->height >> 2);
  XFillRectangle(display,window_info->id,window_info->highlight_context,
    text_info->x,text_info->y,text_info->width,text_info->height);
  XDrawString(display,window_info->id,window_info->annotate_context,x,y,
    text_info->marker,i);
  /*
    Display cursor.
  */
  x+=
    XTextWidth(font_info,text_info->marker,text_info->cursor-text_info->marker);
  XDrawLine(display,window_info->id,window_info->annotate_context,x,y+3,
    x,y-(font_info->ascent+font_info->descent)+3);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X D r a w T r i a n g l e N o r t h                                       %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XDrawTriangleNorth draws a triangle with a highlighted left
%  bevel and a shadowed right and lower bevel.  The highlighted and
%  shadowed bevels create a 3-D effect.
%
%  The format of the XDrawTriangleNorth function is:
%
%      XDrawTriangleNorth(display,window_info,triangle_info)
%
%  A description of each parameter follows:
%
%    o display: Specifies a pointer to the Display structure;  returned from
%      XOpenDisplay.
%
%    o window_info: Specifies a pointer to a X11 XWindowInfo structure.
%
%    o triangle_info: Specifies a pointer to a XWidgetInfo structure.  It
%      contains the extents of the triangle.
%
%
*/
static void XDrawTriangleNorth(display,window_info,triangle_info)
Display
  *display;

XWindowInfo
  *window_info;

XWidgetInfo
  *triangle_info;
{
  int
    x1,
    x2,
    x3,
    y1,
    y2,
    y3;

  unsigned int
    bevel_width;

  XPoint
    points[4];

  /*
    Draw triangle matte.
  */
  x1=triangle_info->x;
  y1=triangle_info->y+triangle_info->height;
  x2=triangle_info->x+(triangle_info->width >> 1);
  y2=triangle_info->y;
  x3=triangle_info->x+triangle_info->width;
  y3=triangle_info->y+triangle_info->height;
  bevel_width=triangle_info->bevel_width;
  points[0].x=x1;
  points[0].y=y1;
  points[1].x=x2;
  points[1].y=y2;
  points[2].x=x3;
  points[2].y=y3;
  XSetMatteColor(display,window_info,triangle_info->raised);
  XFillPolygon(display,window_info->id,window_info->widget_context,points,3,
    Complex,CoordModeOrigin);
  /*
    Draw left bevel.
  */
  points[0].x=x1;
  points[0].y=y1;
  points[1].x=x2;
  points[1].y=y2;
  points[2].x=x2;
  points[2].y=y2-bevel_width-2;
  points[3].x=x1-bevel_width-1;
  points[3].y=y1+bevel_width;
  XSetBevelColor(display,window_info,triangle_info->raised);
  XFillPolygon(display,window_info->id,window_info->widget_context,points,4,
    Complex,CoordModeOrigin);
  /*
    Draw right bevel.
  */
  points[0].x=x2;
  points[0].y=y2;
  points[1].x=x3;
  points[1].y=y3;
  points[2].x=x3+bevel_width;
  points[2].y=y3+bevel_width;
  points[3].x=x2;
  points[3].y=y2-bevel_width;
  XSetBevelColor(display,window_info,(unsigned int) !triangle_info->raised);
  XFillPolygon(display,window_info->id,window_info->widget_context,points,4,
    Complex,CoordModeOrigin);
  /*
    Draw lower bevel.
  */
  points[0].x=x3;
  points[0].y=y3;
  points[1].x=x1;
  points[1].y=y1;
  points[2].x=x1-bevel_width;
  points[2].y=y1+bevel_width;
  points[3].x=x3+bevel_width;
  points[3].y=y3+bevel_width;
  XFillPolygon(display,window_info->id,window_info->widget_context,points,4,
    Complex,CoordModeOrigin);
  XSetFillStyle(display,window_info->widget_context,FillSolid);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X D r a w T r i a n g l e S o u t h                                       %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XDrawTriangleSouth draws a border with a highlighted left and
%  right bevel and a shadowed lower bevel.  The highlighted and shadowed
%  bevels create a 3-D effect.
%
%  The format of the XDrawTriangleSouth function is:
%
%      XDrawTriangleSouth(display,window_info,triangle_info)
%
%  A description of each parameter follows:
%
%    o display: Specifies a pointer to the Display structure;  returned from
%      XOpenDisplay.
%
%    o window_info: Specifies a pointer to a X11 XWindowInfo structure.
%
%    o triangle_info: Specifies a pointer to a XWidgetInfo structure.  It
%      contains the extents of the triangle.
%
%
*/
static void XDrawTriangleSouth(display,window_info,triangle_info)
Display
  *display;

XWindowInfo
  *window_info;

XWidgetInfo
  *triangle_info;
{
  int
    x1,
    x2,
    x3,
    y1,
    y2,
    y3;

  unsigned int
    bevel_width;

  XPoint
    points[4];

  /*
    Draw triangle matte.
  */
  x1=triangle_info->x;
  y1=triangle_info->y;
  x2=triangle_info->x+(triangle_info->width >> 1);
  y2=triangle_info->y+triangle_info->height;
  x3=triangle_info->x+triangle_info->width;
  y3=triangle_info->y;
  bevel_width=triangle_info->bevel_width;
  points[0].x=x1;
  points[0].y=y1;
  points[1].x=x2;
  points[1].y=y2;
  points[2].x=x3;
  points[2].y=y3;
  XSetMatteColor(display,window_info,triangle_info->raised);
  XFillPolygon(display,window_info->id,window_info->widget_context,points,3,
    Complex,CoordModeOrigin);
  /*
    Draw top bevel.
  */
  points[0].x=x3;
  points[0].y=y3;
  points[1].x=x1;
  points[1].y=y1;
  points[2].x=x1-bevel_width;
  points[2].y=y1-bevel_width;
  points[3].x=x3+bevel_width;
  points[3].y=y3-bevel_width;
  XSetBevelColor(display,window_info,triangle_info->raised);
  XFillPolygon(display,window_info->id,window_info->widget_context,points,4,
    Complex,CoordModeOrigin);
  /*
    Draw right bevel.
  */
  points[0].x=x2;
  points[0].y=y2;
  points[1].x=x3+1;
  points[1].y=y3-bevel_width;
  points[2].x=x3+bevel_width;
  points[2].y=y3-bevel_width;
  points[3].x=x2;
  points[3].y=y2+bevel_width;
  XSetBevelColor(display,window_info,(unsigned int) !triangle_info->raised);
  XFillPolygon(display,window_info->id,window_info->widget_context,points,4,
    Complex,CoordModeOrigin);
  /*
    Draw left bevel.
  */
  points[0].x=x1;
  points[0].y=y1;
  points[1].x=x2;
  points[1].y=y2;
  points[2].x=x2;
  points[2].y=y2+bevel_width;
  points[3].x=x1-bevel_width;
  points[3].y=y1-bevel_width;
  XSetBevelColor(display,window_info,triangle_info->raised);
  XFillPolygon(display,window_info->id,window_info->widget_context,points,4,
    Complex,CoordModeOrigin);
  XSetFillStyle(display,window_info->widget_context,FillSolid);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X D r a w W i d g e t T e x t                                             %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XDrawWidgetText first clears the widget and draws a text string
%  justifed left in the x-direction and centered within the y-direction.
%
%  The format of the XDrawWidgetText function is:
%
%      XDrawWidgetText(display,window_info,text_info)
%
%  A description of each parameter follows:
%
%    o display: Specifies a pointer to the Display structure;  returned from
%      XOpenDisplay.
%
%    o window_info: Specifies a pointer to a XWindowText structure.
%
%    o text_info: Specifies a pointer to XWidgetInfo structure.
%
%
*/
void XDrawWidgetText(display,window_info,text_info)
Display
  *display;

XWindowInfo
  *window_info;

XWidgetInfo
  *text_info;
{
  GC
    widget_context;

  int
    x,
    y;

  register char
    *p;

  unsigned int
    height,
    width;

  XFontStruct
    *font_info;

  /*
    Clear the selection.
  */
  widget_context=window_info->annotate_context;
  if (text_info->raised)
    XClearArea(display,window_info->id,text_info->x,text_info->y,
      text_info->width,text_info->height,False);
  else
    {
      XFillRectangle(display,window_info->id,widget_context,text_info->x,
        text_info->y,text_info->width,text_info->height);
      widget_context=window_info->highlight_context;
    }
  if (text_info->text == (char *) NULL)
    return;
  /*
    Insure text will fit within the window.
  */
  font_info=window_info->font_info;
  width=text_info->width-(font_info->max_bounds.width << 1);
  for (p=text_info->text+strlen(text_info->text)-1; p > text_info->text; p--)
    if (XTextWidth(font_info,p,strlen(p)) > width)
      {
	p++;
        break;
      }
  /*
    Draw text.
  */
  height=font_info->ascent+font_info->descent;
  x=text_info->x+font_info->max_bounds.width;
  y=text_info->y+((text_info->height-height) >> 1)+font_info->ascent;
  XDrawString(display,window_info->id,widget_context,x,y,p,strlen(p));
  if (p != text_info->text)
    XDrawLine(display,window_info->id,window_info->annotate_context,
      x,text_info->y,x,text_info->y+text_info->height-1);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X D i a l o g W i d g e t                                                 %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XDialogWidget displays a popup window with a query to the user.
%  The user keys a reply and presses the Ok or Cancel button to exit.  The
%  typed text is returned as the reply function parameter.
%
%  The format of the XDialogWidget routine is:
%
%    XDialogWidget(display,resource_info,window,action,query,reply)
%
%  A description of each parameter follows:
%
%    o display: Specifies a connection to an X server;  returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o action: Specifies a pointer to the action of this widget.
%
%    o query: Specifies a pointer to the query to present to the user.
%
%    o reply: The response from the user is returned in this parameter.
%
%
*/
void XDialogWidget(display,resource_info,window,action,query,reply)
Display
  *display;

XResourceInfo
  *resource_info;

XWindows
  *window;

char
  *action,
  *query,
  *reply;
{
#define CancelButtonText  "Cancel"

  int
    state,
    x,
    y;

  register int
    i;

  unsigned int
    height,
    limit,
    mask,
    width;

  Window
    root_window;

  XEvent
    event;

  XFontStruct
    *font_info;

  XTextProperty
    window_name;

  XWidgetInfo
    action_info,
    cancel_info,
    reply_info,
    text_info;

  XWindowChanges
    window_changes;

  /*
    Determine popup window attributes.
  */
  font_info=window->popup.font_info;
  width=XTextWidth(font_info,action,strlen(action));
  if (XTextWidth(font_info,CancelButtonText,strlen(CancelButtonText)) > width)
    width=XTextWidth(font_info,CancelButtonText,strlen(CancelButtonText));
  width+=(3*font_info->max_bounds.width) >> 1;
  height=font_info->ascent+font_info->descent;
  window->popup.width=width+38*XTextWidth(font_info,"0",1);
  window->popup.height=7*height+3*font_info->max_bounds.width;
  window->popup.min_width=width+19*font_info->max_bounds.width;
  window->popup.min_height=window->popup.height;
  /*
    Initialize button information.
  */
  XGetWidgetInfo(CancelButtonText,&cancel_info);
  cancel_info.width=width;
  cancel_info.height=(3*height) >> 1;
  cancel_info.x=window->popup.width-cancel_info.width-
    ((3*font_info->max_bounds.width) >> 1);
  cancel_info.y=window->popup.height-cancel_info.height-
    ((3*font_info->max_bounds.width) >> 1);
  XGetWidgetInfo(action,&action_info);
  action_info.width=width;
  action_info.height=(3*height) >> 1;
  action_info.x=cancel_info.x-(cancel_info.width+font_info->max_bounds.width+
    (action_info.bevel_width << 1));
  action_info.y=cancel_info.y;
  /*
    Initialize reply information.
  */
  XGetWidgetInfo(reply,&reply_info);
  reply_info.bevel_width--;
  reply_info.width=window->popup.width-(3*font_info->max_bounds.width);
  reply_info.height=height << 1;
  reply_info.x=(3*font_info->max_bounds.width) >> 1;
  reply_info.y=action_info.y-reply_info.height-
    ((3*font_info->max_bounds.width) >> 1);
  /*
    Initialize text information.
  */
  XGetWidgetInfo(query,&text_info);
  text_info.width=reply_info.width+(font_info->max_bounds.width << 1);
  text_info.height=height;
  text_info.x=(font_info->max_bounds.width) >> 1;
  text_info.y=(3*font_info->max_bounds.width) >> 1;
  /*
    Position popup window.
  */
  XQueryPointer(display,XRootWindow(display,window->popup.screen),&root_window,
    &root_window,&x,&y,&window->popup.x,&window->popup.y,&mask);
  window->popup.x-=(reply_info.x+(reply_info.width >> 1));
  limit=XDisplayWidth(display,window->popup.screen)-window->popup.width;
  if (window->popup.x < 0)
    window->popup.x=0;
  else
    if (window->popup.x > limit)
      window->popup.x=limit;
  window->popup.y-=(reply_info.y+(reply_info.height >> 1)-1);
  limit=XDisplayHeight(display,window->popup.screen)-window->popup.height;
  if (window->popup.y < 0)
    window->popup.y=0;
  else
    if (window->popup.y > limit)
      window->popup.y=limit;
  /*
    Map popup window.
  */
  (void) sprintf(window->popup.name,"Dialog");
  (void) XStringListToTextProperty(&window->popup.name,1,&window_name);
  XSetWMName(display,window->popup.id,&window_name);
  window_changes.width=window->popup.width;
  window_changes.height=window->popup.height;
  window_changes.x=window->popup.x;
  window_changes.y=window->popup.y;
  XReconfigureWMWindow(display,window->popup.id,window->popup.screen,CWWidth |
    CWHeight | CWX | CWY,&window_changes);
  XMapRaised(display,window->popup.id);
  XBell(display,0);
  /*
    Respond to X events.
  */
  state=DefaultState;
  XDefineCursor(display,window->image.id,window->image.busy_cursor);
  do
  {
    /*
      Wait for next event.
    */
    XMaskEvent(display,ButtonMotionMask | ButtonPressMask | ButtonReleaseMask |
      ExposureMask | KeyPressMask | StructureNotifyMask,&event);
    switch (event.type)
    {
      case ButtonPress:
      {
        Atom
          type;

        int
          format,
          status;

        unsigned char
          *data;

        unsigned long
          after,
          length;

        if (event.xbutton.window != window->popup.id)
          {
            XBell(display,0);
            break;
          }
        if (MatteIsActive(action_info,event.xbutton))
          {
            /*
              User pressed Ok button.
            */
            action_info.raised=False;
            XDrawBeveledButton(display,&window->popup,&action_info);
            break;
          }
        if (MatteIsActive(cancel_info,event.xbutton))
          {
            /*
              User pressed Cancel button.
            */
            cancel_info.raised=False;
            XDrawBeveledButton(display,&window->popup,&cancel_info);
            break;
          }
        if ((event.xbutton.button == Button3) &&
            (event.xbutton.state & Mod1Mask))
          {
            /*
              Convert Alt-Matte3 to Button2.
            */
            event.xbutton.button=Button2;
            event.xbutton.state&=(~Mod1Mask);
          }
        if (event.xbutton.button != Button2)
          {
            if (MatteIsActive(reply_info,event.xbutton))
              {
                /*
                  Move text cursor to position of button press.
                */
                x=event.xbutton.x-reply_info.x-
                  (font_info->max_bounds.width >> 1);
                for (i=1; i <= (int) strlen(reply_info.marker); i++)
                  if (XTextWidth(font_info,reply_info.marker,i) > x)
                    break;
                reply_info.cursor=reply_info.marker+i-1;
                XDrawMatteText(display,&window->popup,&reply_info);
              }
            break;
          }
        /*
          Obtain response from cut buffer.
        */
        status=XGetWindowProperty(display,XRootWindow(display,0),XA_CUT_BUFFER0,
          0L,2047L,False,XA_STRING,&type,&format,&length,&after,&data);
        if ((status != Success) || (type != XA_STRING) || (format == 32) ||
            (length == 0))
          break;
        if ((strlen(reply_info.text)+length) >= MaxTextLength)
          XBell(display,0);
        else
          {
            /*
              Insert cut buffer in reply text.
            */
            *(data+length)='\0';
            XEditText(display,&reply_info,XK_Insert,(char *) data,state);
            XDrawMatteText(display,&window->popup,&reply_info);
          }
        XFree((void *) data);
        break;
      }
      case ButtonRelease:
      {
        if (!action_info.raised)
          {
            action_info.raised=True;
            XDrawBeveledButton(display,&window->popup,&action_info);
          }
        if (!cancel_info.raised)
          {
            cancel_info.raised=True;
            XDrawBeveledButton(display,&window->popup,&cancel_info);
          }
        if (event.xbutton.window != window->popup.id)
          break;
        if (MatteIsActive(action_info,event.xbutton))
          if (*reply_info.text != '\0')
            state|=ExitState;
          else
            {
              /*
                Do not return if reply is empty.
              */
              XBell(display,0);
              break;
            }
        if (MatteIsActive(cancel_info,event.xbutton))
          {
            *reply_info.text='\0';
            state|=ExitState;
          }
        break;
      }
      case ConfigureNotify:
      {
        /*
          Update widget configuration.
        */
        if (event.xconfigure.window != window->popup.id)
          break;
        if ((window->popup.width == event.xconfigure.width) &&
            (window->popup.height == event.xconfigure.height))
          break;
        window->popup.width=Max(event.xconfigure.width,window->popup.min_width);
        window->popup.height=
          Max(event.xconfigure.height,window->popup.min_height);
        cancel_info.x=window->popup.width-cancel_info.width-
          ((3*font_info->max_bounds.width) >> 1);
        action_info.x=cancel_info.x-(cancel_info.width+
          font_info->max_bounds.width+(action_info.bevel_width << 1));
        reply_info.width=window->popup.width-(3*font_info->max_bounds.width);
        text_info.width=reply_info.width+(font_info->max_bounds.width << 1);
        state|=RedrawWidgetState;
        break;
      }
      case Expose:
      {
        if (event.xexpose.window == window->image.id)
          {
            XRefreshWindow(display,&window->image,&event);
            break;
          }
        if (event.xexpose.window == window->magnify.id)
          if (event.xexpose.count == 0)
            if (window->magnify.mapped)
              {
                XMakeMagnifyImage(display,resource_info,window);
                break;
              }
        if (event.xexpose.window != window->popup.id)
          break;
        if (event.xexpose.count != 0)
          break;
        state|=RedrawWidgetState;
        break;
      }
      case KeyPress:
      {
        static char
          command[MaxTextLength];

        static int
          length;

        static KeySym
          key_symbol;

        if (event.xkey.window != window->popup.id)
          break;
        /*
          Respond to a user key press.
        */
        length=XLookupString((XKeyEvent *) &event.xkey,command,sizeof(command),
          &key_symbol,(XComposeStatus *) NULL);
        *(command+length)='\0';
        if (!MatteIsActive(reply_info,event.xkey))
          break;
        if ((key_symbol == XK_Return) || (key_symbol == XK_KP_Enter))
          {
            if (*reply_info.text != '\0')
              state|=ExitState;
            break;
          }
        if (key_symbol == XK_Control_L)
          {
            state|=ControlState;
            break;
          }
        if (state & ControlState)
          switch (key_symbol)
          {
            case XK_u:
            case XK_U:
            {
              key_symbol=XK_Delete;
              break;
            }
            default:
              break;
          }
        XEditText(display,&reply_info,key_symbol,command,state);
        XDrawMatteText(display,&window->popup,&reply_info);
        break;
      }
      case KeyRelease:
      {
        static char
          command[MaxTextLength];

        static KeySym
          key_symbol;

        if (event.xkey.window != window->popup.id)
          break;
        /*
          Respond to a user key release.
        */
        (void) XLookupString((XKeyEvent *) &event.xkey,command,sizeof(command),
          &key_symbol,(XComposeStatus *) NULL);
        if (key_symbol == XK_Control_L)
          state&=(~ControlState);
        break;
      }
      case MotionNotify:
      {
        if (event.xmotion.window != window->popup.id)
          break;
        if (action_info.raised == MatteIsActive(action_info,event.xmotion))
          {
            /*
              Ok button status changed.
            */
            action_info.raised=!action_info.raised;
            XDrawBeveledButton(display,&window->popup,&action_info);
            break;
          }
        if (cancel_info.raised == MatteIsActive(cancel_info,event.xmotion))
          {
            /*
              Cancel button status changed.
            */
            cancel_info.raised=!cancel_info.raised;
            XDrawBeveledButton(display,&window->popup,&cancel_info);
            break;
          }
        break;
      }
      default:
        break;
    }
    if (state & RedrawWidgetState)
      {
        /*
          Redraw Dialog widget.
        */
        XClearWindow(display,window->popup.id);
        XHighlightWidget(display,&window->popup,5,5);
        XDrawWidgetText(display,&window->popup,&text_info);
        XDrawBeveledMatte(display,&window->popup,&reply_info);
        XDrawMatteText(display,&window->popup,&reply_info);
        XDrawBeveledButton(display,&window->popup,&action_info);
        XDrawBeveledButton(display,&window->popup,&cancel_info);
        state&=(~RedrawWidgetState);
      }
  } while (!(state & ExitState));
  XDefineCursor(display,window->image.id,window->image.cursor);
  XFlush(display);
  XWithdrawWindow(display,window->popup.id,window->popup.screen);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X E d i t T e x t                                                         %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XEditText edits a text string as indicated by the key symbol.
%
%  The format of the XEditText function is:
%
%      XEditText(display,text_info,key_symbol,text,state)
%
%  A description of each parameter follows:
%
%    o display: Specifies a connection to an X server;  returned from
%      XOpenDisplay.
%
%    o text_info: Specifies a pointer to a XWidgetInfo structure.  It
%      contains the extents of the text.
%
%    o key_symbol:  A X11 KeySym that indicates what editing function to
%      perform to the text.
%
%    o text: A character string to insert into the text.
%
%    o state:  An unsigned long that indicates whether the key symbol is a
%      control character or not.
%
%
*/
static void XEditText(display,text_info,key_symbol,text,state)
Display
  *display;

XWidgetInfo
  *text_info;

KeySym
  key_symbol;

char
  *text;

unsigned long
  state;
{
  switch (key_symbol)
  {
    case XK_BackSpace:
    {
      /*
        Erase one character.
      */
      if (text_info->cursor == text_info->text)
        break;
      text_info->cursor--;
      (void) strcpy(text_info->cursor,text_info->cursor+1);
      break;
    }
    case XK_Delete:
    {
      /*
        Erase the entire line of text.
      */
      *text_info->text='\0';
      text_info->cursor=text_info->text;
      break;
    }
    case XK_Left:
    {
      /*
        Move cursor one position left.
      */
      if (text_info->cursor == text_info->text)
        break;
      text_info->cursor--;
      break;
    }
    case XK_Right:
    {
      /*
        Move cursor one position right.
      */
      if (text_info->cursor == (text_info->text+strlen(text_info->text)))
        break;
      text_info->cursor++;
      break;
    }
    default:
    {
      register char
        *p,
        *q;

      register int
        i;

      if (state & ControlState)
        break;
      if (*text == '\0')
        break;
      if (((int) strlen(text_info->text)+1) >= MaxTextLength)
        XBell(display,0);
      else
        {
          /*
            Insert a string into the text.
          */
          q=text_info->text+strlen(text_info->text)+strlen(text);
          for (i=0; i <= (int) strlen(text_info->cursor); i++)
          {
            *q=(*(q-strlen(text)));
            q--;
          }
          p=text;
          for (i=0; i < (int) strlen(text); i++)
            *text_info->cursor++=(*p++);
        }
      break;
    }
  }
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X F i l e B r o w s e r W i d g e t                                       %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XFileBrowserWidget displays a popup window with a file query to the
%  user.  The user keys a reply and presses the Ok or Cancel button to exit.
%  The typed text is returned as the reply function parameter.
%
%  The format of the XFileBrowserWidget routine is:
%
%    XFileBrowserWidget(display,resource_info,window,action,reply)
%
%  A description of each parameter follows:
%
%    o display: Specifies a connection to an X server;  returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o action: Specifies a pointer to the action of this widget.
%
%    o reply: The response from the user is returned in this parameter.
%
%
*/
void XFileBrowserWidget(display,resource_info,window,action,reply)
Display
  *display;

XResourceInfo
  *resource_info;

XWindows
  *window;

char
  *action,
  *reply;
{
#define CancelButtonText  "Cancel"
#define DirectoryText  "Directory:"
#define FilenameText  "File name:"
#define HomeButtonText  "Home"
#define UpButtonText  "Up"

  char
    **filelist,
    home_directory[MaxTextLength],
    text[MaxTextLength],
    working_directory[MaxTextLength];

  int
    number_files,
    state,
    x,
    y;

  register int
    i;

  static char
    glob_pattern[MaxTextLength] = "*";

  unsigned int
    height,
    limit,
    mask,
    visible_filenames,
    width;

  Window
    root_window;

  XEvent
    event;

  XFontStruct
    *font_info;

  XTextProperty
    window_name;

  XWidgetInfo
    action_info,
    cancel_info,
    file_info,
    home_info,
    north_info,
    reply_info,
    scroll_info,
    selection_info,
    slider_info,
    south_info,
    text_info,
    up_info;

  XWindowChanges
    window_changes;

  /*
    Read filelist from current directory.
  */
  XDefineCursor(display,window->image.id,window->image.busy_cursor);
  (void) getcwd(home_directory,MaxTextLength-1);
  (void) strcpy(working_directory,home_directory);
  filelist=ReadDirectory(working_directory,glob_pattern,&number_files);
  /*
    Determine popup window attributes.
  */
  font_info=window->popup.font_info;
  width=XTextWidth(font_info,action,strlen(action));
  if (XTextWidth(font_info,CancelButtonText,strlen(CancelButtonText)) > width)
    width=XTextWidth(font_info,CancelButtonText,strlen(CancelButtonText));
  if (XTextWidth(font_info,HomeButtonText,strlen(HomeButtonText)) > width)
    width=XTextWidth(font_info,HomeButtonText,strlen(HomeButtonText));
  if (XTextWidth(font_info,UpButtonText,strlen(UpButtonText)) > width)
    width=XTextWidth(font_info,UpButtonText,strlen(UpButtonText));
  width+=(3*font_info->max_bounds.width) >> 1;
  if (XTextWidth(font_info,DirectoryText,strlen(DirectoryText)) > width)
    width=XTextWidth(font_info,DirectoryText,strlen(DirectoryText));
  if (XTextWidth(font_info,FilenameText,strlen(FilenameText)) > width)
    width=XTextWidth(font_info,FilenameText,strlen(FilenameText));
  height=font_info->ascent+font_info->descent;
  window->popup.width=width+38*XTextWidth(font_info,"0",1);
  window->popup.height=
    ((93*height) >> 2)+((15*font_info->max_bounds.width) >> 1)+4;
  window->popup.min_width=width+19*font_info->max_bounds.width;
  window->popup.min_height=
    ((23*height) >> 1)+((15*font_info->max_bounds.width) >> 1)+4;
  /*
    Initialize button information.
  */
  XGetWidgetInfo(CancelButtonText,&cancel_info);
  cancel_info.width=width;
  cancel_info.height=(3*height) >> 1;
  cancel_info.x=window->popup.width-cancel_info.width-
    ((3*font_info->max_bounds.width) >> 1)-2;
  cancel_info.y=window->popup.height-cancel_info.height-
    ((3*font_info->max_bounds.width) >> 1);
  XGetWidgetInfo(action,&action_info);
  action_info.width=width;
  action_info.height=(3*height) >> 1;
  action_info.x=cancel_info.x-(cancel_info.width+font_info->max_bounds.width+
    (action_info.bevel_width << 1));
  action_info.y=cancel_info.y;
  XGetWidgetInfo(UpButtonText,&up_info);
  up_info.width=width;
  up_info.height=(3*height) >> 1;
  up_info.x=(3*font_info->max_bounds.width) >> 1;
  up_info.y=((7*font_info->max_bounds.width) >> 1)+height;
  XGetWidgetInfo(HomeButtonText,&home_info);
  home_info.width=width;
  home_info.height=(3*height) >> 1;
  home_info.x=(3*font_info->max_bounds.width) >> 1;
  home_info.y=up_info.y+up_info.height+((3*font_info->max_bounds.width) >> 1);
  /*
    Initialize reply information.
  */
  XGetWidgetInfo(reply,&reply_info);
  reply_info.bevel_width--;
  reply_info.width=window->popup.width-width-
    ((9*font_info->max_bounds.width) >> 1);
  reply_info.height=height << 1;
  reply_info.x=width+3*font_info->max_bounds.width;
  reply_info.y=action_info.y-reply_info.height-
    ((3*font_info->max_bounds.width) >> 1);
  /*
    Initialize scroll information.
  */
  XGetWidgetInfo((char *) NULL,&scroll_info);
  scroll_info.bevel_width--;
  scroll_info.width=height;
  scroll_info.height=reply_info.y-up_info.y-font_info->max_bounds.width;
  scroll_info.x=reply_info.x+(reply_info.width-scroll_info.width);
  scroll_info.y=up_info.y-reply_info.bevel_width;
  scroll_info.raised=False;
  north_info=scroll_info;
  north_info.raised=True;
  north_info.width-=(north_info.bevel_width << 1);
  north_info.height=north_info.width-1;
  north_info.x+=north_info.bevel_width;
  north_info.y+=north_info.bevel_width;
  south_info=north_info;
  south_info.y=scroll_info.y+scroll_info.height-scroll_info.bevel_width-
    south_info.height;
  slider_info=north_info;
  slider_info.id=0;
  slider_info.width-=2;
  slider_info.min_y=north_info.y+north_info.height+north_info.bevel_width+
    slider_info.bevel_width+2;
  slider_info.height=
    scroll_info.height-((slider_info.min_y-scroll_info.y+1) << 1)+2;
  visible_filenames=(scroll_info.height-(height >> 2)-4)/((5*height) >> 2);
  if (number_files > visible_filenames)
    slider_info.height=(visible_filenames*slider_info.height)/number_files;
  slider_info.max_y=south_info.y-south_info.bevel_width-slider_info.height-
    slider_info.bevel_width;
  slider_info.x=scroll_info.x+slider_info.bevel_width+1;
  slider_info.y=slider_info.min_y;
  /*
    Initialize file information.
  */
  XGetWidgetInfo((char *) NULL,&file_info);
  file_info.bevel_width--;
  file_info.width=scroll_info.x-reply_info.x-font_info->max_bounds.width;
  file_info.height=scroll_info.height;
  file_info.x=reply_info.x;
  file_info.y=scroll_info.y;
  /*
    Initialize text information.
  */
  XGetWidgetInfo(text,&text_info);
  text_info.width=reply_info.width+(font_info->max_bounds.width << 1);
  text_info.height=height;
  text_info.x=file_info.x-font_info->max_bounds.width;
  text_info.y=(3*font_info->max_bounds.width) >> 1;
  /*
    Initialize selection information.
  */
  XGetWidgetInfo((char *) NULL,&selection_info);
  selection_info.width=file_info.width;
  selection_info.height=(5*height) >> 2;
  selection_info.x=file_info.x;
  /*
    Position popup window.
  */
  XQueryPointer(display,XRootWindow(display,window->popup.screen),&root_window,
    &root_window,&x,&y,&window->popup.x,&window->popup.y,&mask);
  window->popup.x-=(reply_info.x+(reply_info.width >> 1));
  limit=XDisplayWidth(display,window->popup.screen)-window->popup.width;
  if (window->popup.x < 0)
    window->popup.x=0;
  else
    if (window->popup.x > limit)
      window->popup.x=limit;
  window->popup.y-=(reply_info.y+(reply_info.height >> 1)-1);
  limit=XDisplayHeight(display,window->popup.screen)-window->popup.height;
  if (window->popup.y < 0)
    window->popup.y=0;
  else
    if (window->popup.y > limit)
      window->popup.y=limit;
  /*
    Map popup window.
  */
  (void) sprintf(window->popup.name,"File Browser");
  (void) XStringListToTextProperty(&window->popup.name,1,&window_name);
  XSetWMName(display,window->popup.id,&window_name);
  window_changes.width=window->popup.width;
  window_changes.height=window->popup.height;
  window_changes.x=window->popup.x;
  window_changes.y=window->popup.y;
  XReconfigureWMWindow(display,window->popup.id,window->popup.screen,CWWidth |
    CWHeight | CWX | CWY,&window_changes);
  XMapRaised(display,window->popup.id);
  XBell(display,0);
  /*
    Respond to X events.
  */
  state=DefaultState;
  do
  {
    /*
      Wait for next event.
    */
    XMaskEvent(display,ButtonMotionMask | ButtonPressMask | ButtonReleaseMask |
      ExposureMask | KeyPressMask | StructureNotifyMask,&event);
    switch (event.type)
    {
      case ButtonPress:
      {
        Atom
          type;

        int
          format,
          status;

        unsigned char
          *data;

        unsigned long
          after,
          length;

        if (event.xbutton.window != window->popup.id)
          {
            XBell(display,0);
            break;
          }
        if (MatteIsActive(slider_info,event.xbutton))
          {
            /*
              Track slider.
            */
            slider_info.active=True;
            break;
          }
        if (MatteIsActive(north_info,event.xbutton))
          {
            /*
              Move slider up.
            */
            north_info.raised=False;
            slider_info.id--;
            state|=RedrawFileInfoState;
            break;
          }
        if (MatteIsActive(south_info,event.xbutton))
          {
            /*
              Move slider down.
            */
            south_info.raised=False;
            slider_info.id++;
            state|=RedrawFileInfoState;
            break;
          }
        if (MatteIsActive(scroll_info,event.xbutton))
          {
            /*
              Move slider.
            */
            if (event.xbutton.y < slider_info.y)
              slider_info.id-=visible_filenames;
            else
              slider_info.id+=visible_filenames;
            state|=RedrawFileInfoState;
            break;
          }
        if (MatteIsActive(file_info,event.xbutton))
          {
            unsigned int
              id;

            /*
              User pressed file matte.
            */
            id=slider_info.id+(event.xbutton.y-(file_info.y+(height >> 1))+1)/
              selection_info.height;
            if (id >= number_files)
              break;
            (void) strcpy(reply_info.text,filelist[id]);
            reply_info.marker=reply_info.text;
            reply_info.cursor=reply_info.text+strlen(reply_info.text);
            XDrawMatteText(display,&window->popup,&reply_info);
            if (id == file_info.id)
              {
                (void) strcpy(working_directory,reply_info.text);
                state|=UpdateFilelistState;
              }
            selection_info.id=(~0);
            file_info.id=id;
            state|=RedrawFileInfoState;
            break;
          }
        if (MatteIsActive(up_info,event.xbutton))
          {
            /*
              User pressed Up button.
            */
            up_info.raised=False;
            XDrawBeveledButton(display,&window->popup,&up_info);
            break;
          }
        if (MatteIsActive(home_info,event.xbutton))
          {
            /*
              User pressed Home button.
            */
            home_info.raised=False;
            XDrawBeveledButton(display,&window->popup,&home_info);
            break;
          }
        if (MatteIsActive(action_info,event.xbutton))
          {
            /*
              User pressed action button.
            */
            action_info.raised=False;
            XDrawBeveledButton(display,&window->popup,&action_info);
            break;
          }
        if (MatteIsActive(cancel_info,event.xbutton))
          {
            /*
              User pressed Cancel button.
            */
            cancel_info.raised=False;
            XDrawBeveledButton(display,&window->popup,&cancel_info);
            break;
          }
        if ((event.xbutton.button == Button3) &&
            (event.xbutton.state & Mod1Mask))
          {
            /*
              Convert Alt-Matte3 to Button2.
            */
            event.xbutton.button=Button2;
            event.xbutton.state&=(~Mod1Mask);
          }
        if (event.xbutton.button != Button2)
          {
            if (MatteIsActive(reply_info,event.xbutton))
              {
                /*
                  Move text cursor to position of button press.
                */
                x=event.xbutton.x-reply_info.x-
                  (font_info->max_bounds.width >> 1);
                for (i=1; i <= (int) strlen(reply_info.marker); i++)
                  if (XTextWidth(font_info,reply_info.marker,i) > x)
                    break;
                reply_info.cursor=reply_info.marker+i-1;
                XDrawMatteText(display,&window->popup,&reply_info);
              }
            break;
          }
        /*
          Obtain response from cut buffer.
        */
        status=XGetWindowProperty(display,XRootWindow(display,0),XA_CUT_BUFFER0,
          0L,2047L,False,XA_STRING,&type,&format,&length,&after,&data);
        if ((status != Success) || (type != XA_STRING) || (format == 32) ||
            (length == 0))
          break;
        if ((strlen(reply_info.text)+length) >= MaxTextLength)
          XBell(display,0);
        else
          {
            /*
              Insert cut buffer in reply text.
            */
            *(data+length)='\0';
            XEditText(display,&reply_info,XK_Insert,(char *) data,state);
            XDrawMatteText(display,&window->popup,&reply_info);
          }
        XFree((void *) data);
        break;
      }
      case ButtonRelease:
      {
        if (slider_info.active)
          {
            /*
              Stop tracking slider.
            */
            slider_info.active=False;
            break;
          }
        if (!north_info.raised)
          {
            /*
              User released up button.
            */
            north_info.raised=True;
            XDrawTriangleNorth(display,&window->popup,&north_info);
          }
        if (!south_info.raised)
          {
            /*
              User released down button.
            */
            south_info.raised=True;
            XDrawTriangleSouth(display,&window->popup,&south_info);
          }
        if (!up_info.raised)
          {
            up_info.raised=True;
            XDrawBeveledButton(display,&window->popup,&up_info);
          }
        if (!home_info.raised)
          {
            home_info.raised=True;
            XDrawBeveledButton(display,&window->popup,&home_info);
          }
        if (!action_info.raised)
          {
            action_info.raised=True;
            XDrawBeveledButton(display,&window->popup,&action_info);
          }
        if (!cancel_info.raised)
          {
            cancel_info.raised=True;
            XDrawBeveledButton(display,&window->popup,&cancel_info);
          }
        if (event.xbutton.window != window->popup.id)
          break;
        if (MatteIsActive(up_info,event.xbutton))
          {
            (void) strcpy(working_directory,"..");
            state|=UpdateFilelistState;
          }
        if (MatteIsActive(home_info,event.xbutton))
          {
            (void) strcpy(working_directory,home_directory);
            state|=UpdateFilelistState;
          }
        if (MatteIsActive(action_info,event.xbutton))
          if (*reply_info.text != '\0')
            state|=ExitState;
          else
            {
              /*
                Do not return if reply is empty.
              */
              XBell(display,0);
              break;
            }
        if (MatteIsActive(cancel_info,event.xbutton))
          {
            *reply_info.text='\0';
            state|=ExitState;
          }
        break;
      }
      case ConfigureNotify:
      {
        /*
          Update widget configuration.
        */
        if (event.xconfigure.window != window->popup.id)
          break;
        if ((window->popup.width == event.xconfigure.width) &&
            (window->popup.height == event.xconfigure.height))
          break;
        window->popup.width=Max(event.xconfigure.width,window->popup.min_width);
        window->popup.height=
          Max(event.xconfigure.height,window->popup.min_height);
        cancel_info.x=window->popup.width-cancel_info.width-
          ((3*font_info->max_bounds.width) >> 1);
        cancel_info.y=window->popup.height-cancel_info.height-
          ((3*font_info->max_bounds.width) >> 1);
        action_info.x=cancel_info.x-(cancel_info.width+
          font_info->max_bounds.width+(action_info.bevel_width << 1));
        action_info.y=cancel_info.y;
        reply_info.width=window->popup.width-width-
          ((9*font_info->max_bounds.width) >> 1);
        reply_info.y=action_info.y-reply_info.height-
          ((3*font_info->max_bounds.width) >> 1);
        scroll_info.height=reply_info.y-up_info.y-font_info->max_bounds.width;
        scroll_info.x=reply_info.x+reply_info.width-scroll_info.width;
        north_info.x=scroll_info.x+north_info.bevel_width;
        south_info.x=scroll_info.x+south_info.bevel_width;
        south_info.y=scroll_info.y+scroll_info.height-scroll_info.bevel_width-
          south_info.height;
        slider_info.x=scroll_info.x+slider_info.bevel_width+1;
        slider_info.height=
          scroll_info.height-((slider_info.min_y-scroll_info.y+1) << 1)+2;
        visible_filenames=
          (scroll_info.height-(height >> 2)-4)/((5*height) >> 2);
        if (number_files > visible_filenames)
          slider_info.height=
            (visible_filenames*slider_info.height)/number_files;
        slider_info.max_y=south_info.y-south_info.bevel_width-
          slider_info.height-slider_info.bevel_width;
        file_info.width=scroll_info.x-reply_info.x-font_info->max_bounds.width;
        file_info.height=reply_info.y-up_info.y-font_info->max_bounds.width;
        text_info.width=file_info.width+(font_info->max_bounds.width << 1);
        selection_info.width=file_info.width;
        state|=RedrawWidgetState;
        break;
      }
      case Expose:
      {
        if (event.xexpose.window == window->image.id)
          {
            XRefreshWindow(display,&window->image,&event);
            break;
          }
        if (event.xexpose.window == window->magnify.id)
          if (event.xexpose.count == 0)
            if (window->magnify.mapped)
              {
                XMakeMagnifyImage(display,resource_info,window);
                break;
              }
        if (event.xexpose.window != window->popup.id)
          break;
        if (event.xexpose.count != 0)
          break;
        state|=RedrawWidgetState;
        break;
      }
      case KeyPress:
      {
        static char
          command[MaxTextLength];

        static int
          length;

        static KeySym
          key_symbol;

        if (event.xkey.window != window->popup.id)
          break;
        /*
          Respond to a user key press.
        */
        length=XLookupString((XKeyEvent *) &event.xkey,command,sizeof(command),
          &key_symbol,(XComposeStatus *) NULL);
        *(command+length)='\0';
        if (MatteIsActive(slider_info,event.xbutton))
          break;
        if (MatteIsActive(north_info,event.xkey))
          {
            /*
              Move slider up.
            */
            slider_info.id--;
            state|=RedrawFileInfoState;
            break;
          }
        if (MatteIsActive(south_info,event.xkey))
          {
            /*
              Move slider down.
            */
            slider_info.id++;
            state|=RedrawFileInfoState;
            break;
          }
        if (MatteIsActive(scroll_info,event.xkey))
          {
            /*
              Move slider.
            */
            if (event.xkey.y < slider_info.y)
              slider_info.id-=visible_filenames;
            else
              slider_info.id+=visible_filenames;
            state|=RedrawFileInfoState;
            break;
          }
        if (!MatteIsActive(reply_info,event.xkey))
          break;
        if ((key_symbol == XK_Return) || (key_symbol == XK_KP_Enter))
          {
            /*
              Read new directory or glob patterm.
            */
            if (*reply_info.text == '\0')
              break;
            if ((strchr(reply_info.text,'*') != (char *) NULL) ||
                (strchr(reply_info.text,'[') != (char *) NULL) ||
                (strchr(reply_info.text,']') != (char *) NULL))
              (void) strcpy(glob_pattern,reply_info.text);
            else
              (void) strcpy(working_directory,reply_info.text);
            state|=UpdateFilelistState;
            break;
          }
        if (key_symbol == XK_Control_L)
          {
            state|=ControlState;
            break;
          }
        if (state & ControlState)
          switch (key_symbol)
          {
            case XK_u:
            case XK_U:
            {
              key_symbol=XK_Delete;
              break;
            }
            default:
              break;
          }
        XEditText(display,&reply_info,key_symbol,command,state);
        XDrawMatteText(display,&window->popup,&reply_info);
        break;
      }
      case KeyRelease:
      {
        static char
          command[MaxTextLength];

        static KeySym
          key_symbol;

        if (event.xkey.window != window->popup.id)
          break;
        /*
          Respond to a user key release.
        */
        (void) XLookupString((XKeyEvent *) &event.xkey,command,sizeof(command),
          &key_symbol,(XComposeStatus *) NULL);
        if (key_symbol == XK_Control_L)
          state&=(~ControlState);
        break;
      }
      case MotionNotify:
      {
        if (event.xmotion.window != window->popup.id)
          break;
        if (slider_info.active)
          {
            /*
              Move slider matte.
            */
            slider_info.y=event.xmotion.y-
              ((slider_info.height+slider_info.bevel_width) >> 1)+1;
            if (slider_info.y < slider_info.min_y)
              slider_info.y=slider_info.min_y;
            if (slider_info.y > slider_info.max_y)
              slider_info.y=slider_info.max_y;
            slider_info.id=(number_files*(slider_info.y-slider_info.min_y+1))/
              (slider_info.max_y-slider_info.min_y+1);
            state|=RedrawFileInfoState;
            break;
          }
        if (north_info.raised == MatteIsActive(north_info,event.xmotion))
          {
            /*
              Up triangle status changed.
            */
            north_info.raised=!north_info.raised;
            break;
          }
        if (south_info.raised == MatteIsActive(south_info,event.xmotion))
          {
            /*
              Down triangle status changed.
            */
            south_info.raised=!south_info.raised;
            break;
          }
        if (up_info.raised == MatteIsActive(up_info,event.xmotion))
          {
            /*
              Up button status changed.
            */
            up_info.raised=!up_info.raised;
            XDrawBeveledButton(display,&window->popup,&up_info);
            break;
          }
        if (home_info.raised == MatteIsActive(home_info,event.xmotion))
          {
            /*
              Home button status changed.
            */
            home_info.raised=!home_info.raised;
            XDrawBeveledButton(display,&window->popup,&home_info);
            break;
          }
        if (action_info.raised == MatteIsActive(action_info,event.xmotion))
          {
            /*
              Ok button status changed.
            */
            action_info.raised=!action_info.raised;
            XDrawBeveledButton(display,&window->popup,&action_info);
            break;
          }
        if (cancel_info.raised == MatteIsActive(cancel_info,event.xmotion))
          {
            /*
              Cancel button status changed.
            */
            cancel_info.raised=!cancel_info.raised;
            XDrawBeveledButton(display,&window->popup,&cancel_info);
            break;
          }
        break;
      }
      default:
        break;
    }
    if (state & RedrawWidgetState)
      {
        /*
          Redraw file browser window.
        */
        XClearWindow(display,window->popup.id);
        XHighlightWidget(display,&window->popup,5,5);
        x=(3*font_info->max_bounds.width) >> 1;
        y=text_info.y+((text_info.height-height) >> 1)+font_info->ascent;
        XDrawString(display,window->popup.id,window->popup.annotate_context,
	  x,y,DirectoryText,strlen(DirectoryText));
        (void) sprintf(text_info.text,"%s/%s",working_directory,glob_pattern);
        XDrawWidgetText(display,&window->popup,&text_info);
        XDrawBeveledButton(display,&window->popup,&up_info);
        XDrawBeveledButton(display,&window->popup,&home_info);
        XDrawBeveledMatte(display,&window->popup,&file_info);
        x=(3*font_info->max_bounds.width) >> 1;
        y=reply_info.y+((reply_info.height-height) >> 1)+font_info->ascent;
        XDrawString(display,window->popup.id,window->popup.annotate_context,
	  x,y,FilenameText,strlen(FilenameText));
        XDrawBeveledMatte(display,&window->popup,&reply_info);
        XDrawMatteText(display,&window->popup,&reply_info);
        XDrawBeveledButton(display,&window->popup,&action_info);
        XDrawBeveledButton(display,&window->popup,&cancel_info);
        selection_info.id=(~0);
        state|=RedrawFileInfoState;
        state&=(~RedrawWidgetState);
      }
    if (state & UpdateFilelistState)
      {
        /*
          Update file list.
        */
        if (number_files > 0)
          {
            /*
              Free file list.
            */
            for (i=0; i < number_files; i++)
              (void) free((char *) filelist[i]);
            (void) free((char *) filelist);
          }
        filelist=ReadDirectory(working_directory,glob_pattern,&number_files);
        if (filelist == (char **) NULL)
          {
            /*
              Reply is a filename, exit.
            */
            action_info.raised=False;
            XDrawBeveledButton(display,&window->popup,&action_info);
            break;
          }
        slider_info.height=
          scroll_info.height-((slider_info.min_y-scroll_info.y+1) << 1)+1;
        if (number_files > visible_filenames)
          slider_info.height=
            (visible_filenames*slider_info.height)/number_files;
        slider_info.max_y=south_info.y-south_info.bevel_width-
          slider_info.height-slider_info.bevel_width-1;
        slider_info.id=0;
        selection_info.id=(~0);
        file_info.id=(~0);
        state|=RedrawFileInfoState;
        /*
          Redraw directory name & reply.
        */
        *reply_info.text='\0';
        reply_info.cursor=reply_info.text;
        (void) sprintf(text_info.text,"%s/%s",working_directory,glob_pattern);
        XDrawWidgetText(display,&window->popup,&text_info);
        XDrawMatteText(display,&window->popup,&reply_info);
        XHighlightWidget(display,&window->popup,5,5);
        state&=(~UpdateFilelistState);
      }
    if (state & RedrawFileInfoState)
      {
        /*
          Determine slider id and position.
        */
        if (slider_info.id >= number_files)
          slider_info.id=number_files-1;
        if ((slider_info.id < 0) || (number_files <= visible_filenames))
          slider_info.id=0;
        slider_info.y=slider_info.min_y;
        if (number_files > 0)
          slider_info.y+=
            slider_info.id*(slider_info.max_y-slider_info.min_y+1)/number_files;
        if (slider_info.id != selection_info.id)
          {
            /*
              Redraw scroll bar and file names.
            */
            selection_info.id=slider_info.id;
            selection_info.y=file_info.y+(height >> 3)+2;
            for (i=0; i < visible_filenames; i++)
            {
              selection_info.raised=(slider_info.id+i) != file_info.id;
              selection_info.text=(char *) NULL;
              if ((slider_info.id+i) < number_files)
                selection_info.text=filelist[slider_info.id+i];
              XDrawWidgetText(display,&window->popup,&selection_info);
              selection_info.y+=(int) selection_info.height;
            }
            XDrawBeveledButton(display,&window->popup,&scroll_info);
            XDrawTriangleNorth(display,&window->popup,&north_info);
            XDrawTriangleSouth(display,&window->popup,&south_info);
            XDrawBeveledButton(display,&window->popup,&slider_info);
          }
        state&=(~RedrawFileInfoState);
      }
  } while (!(state & ExitState));
  XDefineCursor(display,window->image.id,window->image.cursor);
  XFlush(display);
  XWithdrawWindow(display,window->popup.id,window->popup.screen);
  if (number_files > 0)
    {
      /*
        Free file list.
      */
      for (i=0; i < number_files; i++)
        (void) free((char *) filelist[i]);
      (void) free((char *) filelist);
    }
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X G e t W i d g e t I n f o                                               %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XGetWidgetInfo initializes the XWidgetInfo structure.
%
%  The format of the XGetWidgetInfo function is:
%
%      XHighlightWidget(text,widget_info)
%
%  A description of each parameter follows:
%
%    o text: A string of characters associated with the widget.
%
%    o widget_info: Specifies a pointer to a X11 XWidgetInfo structure.
%
%
*/
static void XGetWidgetInfo(text,widget_info)
char
  *text;

XWidgetInfo
  *widget_info;
{
  /*
    Initialize widget info.
  */
   widget_info->id=(~0);
   widget_info->bevel_width=3;
   widget_info->width=1;
   widget_info->height=1;
   widget_info->x=0;
   widget_info->y=0;
   widget_info->min_y=0;
   widget_info->max_y=0;
   widget_info->raised=True;
   widget_info->active=False;
   widget_info->text=text;
   widget_info->cursor=text;
   if (text != (char *) NULL)
     widget_info->cursor+=strlen(text);
   widget_info->marker=text;
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X H i g h l i g h t W i d g e t                                           %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XHighlightWidget draws a highlighted border around a window.
%
%  The format of the XHighlightWidget function is:
%
%      XHighlightWidget(display,window_info,x,y)
%
%  A description of each parameter follows:
%
%    o display: Specifies a pointer to the Display structure;  returned from
%      XOpenDisplay.
%
%    o window_info: Specifies a pointer to a X11 XWindowInfo structure.
%
%    o x: Specifies an integer representing the rectangle offset in the
%      x-direction.
%
%    o y: Specifies an integer representing the rectangle offset in the
%      y-direction.
%
%
*/
static void XHighlightWidget(display,window_info,x,y)
Display
  *display;

XWindowInfo
  *window_info;

int
  x,
  y;
{
  /*
    Draw the widget highlighting rectangle.
  */
  XSetBevelColor(display,window_info,True);
  XDrawRectangle(display,window_info->id,window_info->widget_context,x,y,
    window_info->width-(x << 1),window_info->height-(y << 1));
  XDrawRectangle(display,window_info->id,window_info->widget_context,x-1,y-1,
    window_info->width-(x << 1)+1,window_info->height-(y << 1)+1);
  XSetBevelColor(display,window_info,False);
  XDrawRectangle(display,window_info->id,window_info->widget_context,x-1,y-1,
    window_info->width-(x << 1),window_info->height-(y << 1));
  XSetFillStyle(display,window_info->widget_context,FillSolid);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X M e n u W i d g e t                                                     %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XMenuWidget maps a menu and returns the command pointed to by the
%  user when the button is released.
%
%  The format of the XMenuWidget routine is:
%
%    selection_number=XMenuWidget(display,window,x,y,menu_title,menu_selections,
%      number_selections,item)
%
%  A description of each parameter follows:
%
%    o selection_number: Specifies the number of the selection that the
%      user choose.
%
%    o display: Specifies a connection to an X server;  returned from
%      XOpenDisplay.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o x: Specifies an unsigned integer representing the root offset in the
%      x-direction.
%
%    o y: Specifies an unsigned integer representing the root offset in the
%      y-direction.
%
%    o menu_title: Specifies a character string that describes the menu
%      selections.
%
%    o menu_selections: Specifies a pointer to one or more strings that
%      make up the choices in the menu.
%
%    o number_selections: Specifies the number of choices in the menu.
%
%    o item: Specifies a character array.  The item selected from the menu
%      is returned here.
%
%
*/
int XMenuWidget(display,window,x,y,menu_title,menu_selections,number_selections,
  item)
Display
  *display;

XWindows
  *window;

int
  x,
  y;

char
  *menu_title,
  **menu_selections;

unsigned int
  number_selections;

char
  *item;
{
  Cursor
    cursor;

  int
    id;

  unsigned int
    height,
    limit,
    title_height,
    width;

  unsigned long
    state;

  XEvent
    event;

  XFontStruct
    *font_info;

  XSetWindowAttributes
    window_attributes;

  XWidgetInfo
    selection_info;

  XWindowChanges
    window_changes;

  /*
    Determine popup window attributes.
  */
  font_info=window->popup.font_info;
  window->popup.width=XTextWidth(font_info,menu_title,strlen(menu_title));
  for (id=0; id < number_selections; id++)
  {
    width=XTextWidth(font_info,menu_selections[id],strlen(menu_selections[id]));
    if (width > window->popup.width)
      window->popup.width=width;
  }
  window->popup.width+=(font_info->max_bounds.width << 1);
  title_height=(font_info->descent+font_info->ascent) << 1;
  height=(5*(font_info->ascent+font_info->descent)) >> 2;
  window->popup.height=title_height+number_selections*height+(height >> 2);
  /*
    Position popup window.
  */
  window->popup.x=x-font_info->max_bounds.width;
  limit=XDisplayWidth(display,window->popup.screen)-window->popup.width;
  if (window->popup.x < 0)
    window->popup.x=0;
  else
    if (window->popup.x > limit)
      window->popup.x=limit;
  window->popup.y=y-((3*title_height) >> 2);
  limit=XDisplayHeight(display,window->popup.screen)-window->popup.height;
  if (window->popup.y < 0)
    window->popup.y=0;
  else
    if (window->popup.y > limit)
      window->popup.y=limit;
  /*
    Map popup window.
  */
  window_attributes.override_redirect=True;
  XChangeWindowAttributes(display,window->popup.id,CWOverrideRedirect,
    &window_attributes);
  window_changes.width=window->popup.width;
  window_changes.height=window->popup.height;
  window_changes.x=window->popup.x;
  window_changes.y=window->popup.y;
  XReconfigureWMWindow(display,window->popup.id,window->popup.screen,CWWidth |
    CWHeight | CWX | CWY,&window_changes);
  XMapRaised(display,window->popup.id);
  /*
    Draw title.
  */
  width=XTextWidth(font_info,menu_title,strlen(menu_title));
  x=(window->popup.width-width) >> 1;
  y=font_info->ascent+(title_height >> 2);
  XDrawString(display,window->popup.id,window->popup.annotate_context,x,y,
    menu_title,strlen(menu_title));
  /*
    Draw highlighting line.
  */
  XSetBevelColor(display,&window->popup,False);
  XDrawLine(display,window->popup.id,window->popup.widget_context,0,
    (int) (title_height-2),(int) window->popup.width,(int) (title_height-2));
  XSetBevelColor(display,&window->popup,True);
  XDrawLine(display,window->popup.id,window->popup.widget_context,0,
    (int) (title_height-1),(int) window->popup.width,(int) (title_height-1));
  XSetFillStyle(display,window->popup.widget_context,FillSolid);
  /*
    Draw menu selections.
  */
  XGetWidgetInfo((char *) NULL,&selection_info);
  selection_info.width=window->popup.width;
  selection_info.height=height;
  selection_info.x=0;
  selection_info.y=title_height;
  for (id=0; id < number_selections; id++)
  {
    selection_info.text=menu_selections[id];
    XDrawWidgetText(display,&window->popup,&selection_info);
    selection_info.y+=(int) selection_info.height;
  }
  /*
    Respond to X events.
  */
  cursor=XCreateFontCursor(display,XC_right_ptr);
  XDefineCursor(display,window->image.id,cursor);
  state=DefaultState;
  do
  {
    /*
      Wait for next event.
    */
    XMaskEvent(display,ButtonMotionMask | ButtonPressMask | ButtonReleaseMask |
      EnterWindowMask | LeaveWindowMask,&event);
    switch (event.type)
    {
      case ButtonPress:
        break;
      case ButtonRelease:
      {
        /*
          Exit menu.
        */
        *item='\0';
        state|=ExitState;
        break;
      }
      case EnterNotify:
      {
        if (event.xcrossing.window != window->popup.id)
          break;
        id=((event.xcrossing.y-title_height)/(int) selection_info.height);
        if ((id < 0) || (id >= number_selections))
          break;
        /*
          Highlight this selection.
        */
        selection_info.id=id;
        selection_info.y=title_height+selection_info.id*selection_info.height;
        selection_info.raised=False;
        selection_info.text=menu_selections[selection_info.id];
        XDrawWidgetText(display,&window->popup,&selection_info);
        break;
      }
      case LeaveNotify:
      {
        if (event.xcrossing.window != window->popup.id)
          break;
        id=selection_info.id;
        if ((id < 0) || (id >= number_selections))
          break;
        /*
          Unhighlight last selection.
        */
        selection_info.id=(~0);
        selection_info.raised=True;
        selection_info.text=menu_selections[id];
        XDrawWidgetText(display,&window->popup,&selection_info);
        break;
      }
      case MotionNotify:
      {
        if (event.xmotion.window != window->popup.id)
          break;
        /*
          Determine if pointer has moved to a new selection.
        */
        id=(event.xmotion.y-title_height)/(int) selection_info.height;
        if ((selection_info.id >= 0) && (selection_info.id < number_selections))
          {
            /*
              Unhighlight last selection.
            */
            if (id == selection_info.id)
              break;
            selection_info.raised=True;
            selection_info.text=menu_selections[selection_info.id];
            XDrawWidgetText(display,&window->popup,&selection_info);
          }
        selection_info.id=id;
        if ((id < 0) || (id >= number_selections))
          break;
        /*
          Highlight this selection.
        */
        selection_info.y=title_height+selection_info.id*selection_info.height;
        selection_info.raised=False;
        selection_info.text=menu_selections[selection_info.id];
        XDrawWidgetText(display,&window->popup,&selection_info);
        break;
      }
      default:
        break;
    }
  } while (!(state & ExitState));
  XDefineCursor(display,window->image.id,window->image.cursor);
  window_attributes.override_redirect=False;
  XChangeWindowAttributes(display,window->popup.id,CWOverrideRedirect,
    &window_attributes);
  XFlush(display);
  XWithdrawWindow(display,window->popup.id,window->popup.screen);
  if ((selection_info.id < 0) || (selection_info.id >= number_selections))
    return(~0);
  (void) strcpy(item,menu_selections[selection_info.id]);
  return(selection_info.id);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X N o t i c e W i d g e t                                                 %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XNoticeWidget displays a popup window with a notice to the user.
%  The function returns when the user presses the "Dismiss" button.
%
%  The format of the XNoticeWidget routine is:
%
%    XNoticeWidget(display,resource_info,window,message,qualifier)
%
%  A description of each parameter follows:
%
%    o display: Specifies a connection to an X server;  returned from
%      XOpenDisplay.
%
%    o resource_info: Specifies a pointer to a X11 XResourceInfo structure.
%
%    o window: Specifies a pointer to a XWindows structure.
%
%    o message: Specifies the message to display before terminating the
%      program.
%
%    o qualifier: Specifies any qualifier to the message.
%
%
*/
void XNoticeWidget(display,resource_info,window,message,qualifier)
Display
  *display;

XResourceInfo
  *resource_info;

XWindows
  *window;

char
  *message,
  *qualifier;
{
#define DismissButtonText  "Dismiss"

  int
    limit,
    state,
    x,
    y;

  unsigned int
    height,
    mask,
    width;

  Window
    root_window;

  XEvent
    event;

  XFontStruct
    *font_info;

  XTextProperty
    window_name;

  XWidgetInfo
    dismiss_info;

  XWindowChanges
    window_changes;

  /*
    Determine popup window attributes.
  */
  font_info=window->popup.font_info;
  width=XTextWidth(font_info,DismissButtonText,strlen(DismissButtonText));
  if (message != (char *) NULL)
    if (XTextWidth(font_info,message,strlen(message)) > width)
      width=XTextWidth(font_info,message,strlen(message));
  if (qualifier != (char *) NULL)
    if (XTextWidth(font_info,qualifier,strlen(qualifier)) > width)
      width=XTextWidth(font_info,qualifier,strlen(qualifier));
  height=(font_info->ascent+font_info->descent);
  window->popup.width=width+12*font_info->max_bounds.width;
  window->popup.height=12*height;
  window->popup.min_width=width+(font_info->max_bounds.width << 1);
  window->popup.min_height=7*height;
  /*
    Initialize Dismiss button information.
  */
  XGetWidgetInfo(DismissButtonText,&dismiss_info);
  dismiss_info.width=((3*font_info->max_bounds.width) >> 1)+
    XTextWidth(font_info,DismissButtonText,strlen(DismissButtonText));
  dismiss_info.height=(3*height) >> 1;
  dismiss_info.x=(window->popup.width >> 1)-(dismiss_info.width >> 1);
  dismiss_info.y=window->popup.height-(dismiss_info.height << 1);
  /*
    Position popup window.
  */
  XQueryPointer(display,XRootWindow(display,window->popup.screen),&root_window,
    &root_window,&x,&y,&window->popup.x,&window->popup.y,&mask);
  window->popup.x-=(dismiss_info.x+(dismiss_info.width >> 1));
  limit=XDisplayWidth(display,window->popup.screen)-window->popup.width;
  if (window->popup.x < 0)
    window->popup.x=0;
  else
    if (window->popup.x > limit)
      window->popup.x=limit;
  window->popup.y-=dismiss_info.y-((3*height) >> 1);
  limit=XDisplayHeight(display,window->popup.screen)-window->popup.height;
  if (window->popup.y < 0)
    window->popup.y=0;
  else
    if (window->popup.y > limit)
      window->popup.y=limit;
  /*
    Map popup window.
  */
  (void) sprintf(window->popup.name,"Notice");
  (void) XStringListToTextProperty(&window->popup.name,1,&window_name);
  XSetWMName(display,window->popup.id,&window_name);
  window_changes.width=window->popup.width;
  window_changes.height=window->popup.height;
  window_changes.x=window->popup.x;
  window_changes.y=window->popup.y;
  XReconfigureWMWindow(display,window->popup.id,window->popup.screen,CWWidth |
    CWHeight | CWX | CWY,&window_changes);
  XMapRaised(display,window->popup.id);
  XBell(display,0);
  /*
    Respond to X events.
  */
  state=DefaultState;
  XDefineCursor(display,window->image.id,window->image.busy_cursor);
  do
  {
    /*
      Wait for next event.
    */
    XMaskEvent(display,ButtonMotionMask | ButtonPressMask | ButtonReleaseMask |
      ExposureMask | KeyPressMask | StructureNotifyMask,&event);
    switch (event.type)
    {
      case ButtonPress:
      {
        if (event.xbutton.window != window->popup.id)
          {
            XBell(display,0);
            break;
          }
        if (MatteIsActive(dismiss_info,event.xbutton))
          {
            /*
              User pressed Dismiss button.
            */
            dismiss_info.raised=False;
            XDrawBeveledButton(display,&window->popup,&dismiss_info);
            break;
          }
        break;
      }
      case ButtonRelease:
      {
        if (!dismiss_info.raised)
          {
            dismiss_info.raised=True;
            XDrawBeveledButton(display,&window->popup,&dismiss_info);
          }
        if (event.xbutton.window != window->popup.id)
          break;
        if (MatteIsActive(dismiss_info,event.xbutton))
          state|=ExitState;
        break;
      }
      case ConfigureNotify:
      {
        /*
          Update widget configuration.
        */
        if (event.xconfigure.window != window->popup.id)
          break;
        if ((window->popup.width == event.xconfigure.width) &&
            (window->popup.height == event.xconfigure.height))
          break;
        window->popup.width=Max(event.xconfigure.width,window->popup.min_width);
        window->popup.height=
          Max(event.xconfigure.height,window->popup.min_height);
        dismiss_info.x=(window->popup.width >> 1)-(dismiss_info.width >> 1);
        dismiss_info.y=window->popup.height-(dismiss_info.height << 1);
        state|=RedrawWidgetState;
        break;
      }
      case Expose:
      {
        if (event.xexpose.window == window->image.id)
          {
            XRefreshWindow(display,&window->image,&event);
            break;
          }
        if (event.xexpose.window == window->magnify.id)
          if (event.xexpose.count == 0)
            if (window->magnify.mapped)
              {
                XMakeMagnifyImage(display,resource_info,window);
                break;
              }
        if (event.xexpose.window != window->popup.id)
          break;
        if (event.xexpose.count != 0)
          break;
        state|=RedrawWidgetState;
        break;
      }
      case KeyPress:
      {
        static char
          command[MaxTextLength];

        static KeySym
          key_symbol;

        if (event.xkey.window != window->popup.id)
          break;
        /*
          Respond to a user key press.
        */
        (void) XLookupString((XKeyEvent *) &event.xkey,command,sizeof(command),
          &key_symbol,(XComposeStatus *) NULL);
        if ((key_symbol == XK_Return) || (key_symbol == XK_KP_Enter))
          {
            dismiss_info.raised=False;
            XDrawBeveledButton(display,&window->popup,&dismiss_info);
            state|=ExitState;
            break;
          }
        break;
      }
      case MotionNotify:
      {
        if (event.xmotion.window != window->popup.id)
          break;
        if (dismiss_info.raised == MatteIsActive(dismiss_info,event.xmotion))
          {
            /*
              Dismiss button status changed.
            */
            dismiss_info.raised=!dismiss_info.raised;
            XDrawBeveledButton(display,&window->popup,&dismiss_info);
            break;
          }
        break;
      }
      default:
        break;
    }
    if (state & RedrawWidgetState)
      {
        /*
          Redraw Notice widget.
        */
        XClearWindow(display,window->popup.id);
        XHighlightWidget(display,&window->popup,5,5);
        width=XTextWidth(font_info,message,strlen(message));
        x=(window->popup.width >> 1)-(width >> 1);
        y=(window->popup.height >> 1)-(height << 1);
        XDrawString(display,window->popup.id,window->popup.annotate_context,x,y,
          message,strlen(message));
        if (qualifier != (char *) NULL)
          {
            width=XTextWidth(font_info,qualifier,strlen(qualifier));
            x=(window->popup.width >> 1)-(width >> 1);
            y+=height;
            XDrawString(display,window->popup.id,window->popup.annotate_context,
              x,y,qualifier,strlen(qualifier));
          }
        XDrawBeveledButton(display,&window->popup,&dismiss_info);
        state&=(~RedrawWidgetState);
      }
  } while (!(state & ExitState));
  XDefineCursor(display,window->image.id,window->image.cursor);
  XFlush(display);
  XWithdrawWindow(display,window->popup.id,window->popup.screen);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X S e t B e v e l C o l o r                                               %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XSetBevelColor sets the graphic context for drawing a beveled
%  border.
%
%  The format of the XSetBevelColor function is:
%
%      XSetBevelColor(display,window_info,raised)
%
%  A description of each parameter follows:
%
%    o display: Specifies a pointer to the Display structure;  returned from
%      XOpenDisplay.
%
%    o window_info: Specifies a pointer to a X11 XWindowInfo structure.
%
%    o raised: A value other than zero indicates the color show be a
%      "highlight" color, otherwise the "shadow" color is set.
%
%
*/
static void XSetBevelColor(display,window_info,raised)
Display
  *display;

XWindowInfo
  *window_info;

unsigned int
  raised;
{
  if (window_info->depth == 1)
    {
      Pixmap
        stipple;

      /*
        Monochrome window.
      */
      XSetBackground(display,window_info->widget_context,
        XBlackPixel(display,window_info->screen));
      XSetForeground(display,window_info->widget_context,
        XWhitePixel(display,window_info->screen));
      XSetFillStyle(display,window_info->widget_context,FillOpaqueStippled);
      stipple=window_info->highlight_stipple;
      if (!raised)
        stipple=window_info->shadow_stipple;
      XSetStipple(display,window_info->widget_context,stipple);
    }
  else
    if (raised)
      XSetForeground(display,window_info->widget_context,
        window_info->pixel_info->highlight_color.pixel);
    else
      XSetForeground(display,window_info->widget_context,
        window_info->pixel_info->shadow_color.pixel);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X S e t M a t t e C o l o r                                               %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XSetMatteColor sets the graphic context for drawing the matte.
%
%  The format of the XSetMatteColor function is:
%
%      XSetMatteColor(display,window_info,raised)
%
%  A description of each parameter follows:
%
%    o display: Specifies a pointer to the Display structure;  returned from
%      XOpenDisplay.
%
%    o window_info: Specifies a pointer to a X11 XWindowInfo structure.
%
%    o raised: A value other than zero indicates the matte is active.
%
%
*/
static void XSetMatteColor(display,window_info,raised)
Display
  *display;

XWindowInfo
  *window_info;

unsigned int
  raised;
{
  if (window_info->depth == 1)
    {
      /*
        Monochrome window.
      */
      if (raised)
        XSetForeground(display,window_info->widget_context,
          XWhitePixel(display,window_info->screen));
      else
        XSetForeground(display,window_info->widget_context,
          XBlackPixel(display,window_info->screen));
    }
  else
    if (raised)
      XSetForeground(display,window_info->widget_context,
        window_info->pixel_info->matte_color.pixel);
    else
      XSetForeground(display,window_info->widget_context,
        window_info->pixel_info->depth_color.pixel);
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                                                                             %
%                                                                             %
%   X S e t T e x t C o l o r                                                 %
%                                                                             %
%                                                                             %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Function XSetTextColor sets the graphic context for drawing text on a
%  matte.
%
%  The format of the XSetTextColor function is:
%
%      XSetTextColor(display,window_info,raised)
%
%  A description of each parameter follows:
%
%    o display: Specifies a pointer to the Display structure;  returned from
%      XOpenDisplay.
%
%    o window_info: Specifies a pointer to a X11 XWindowInfo structure.
%
%    o raised: A value other than zero indicates the color show be a
%      "highlight" color, otherwise the "shadow" color is set.
%
%
*/
static void XSetTextColor(display,window_info,raised)
Display
  *display;

XWindowInfo
  *window_info;

unsigned int
  raised;
{
  if (window_info->depth == 1)
    {
      /*
        Monochrome window.
      */
      if (raised)
        XSetForeground(display,window_info->widget_context,
          XBlackPixel(display,window_info->screen));
      else
        XSetForeground(display,window_info->widget_context,
          XWhitePixel(display,window_info->screen));
    }
  else
    XSetForeground(display,window_info->widget_context,
      window_info->pixel_info->foreground_color.pixel);
}
